/*
 * OCERA NETWORK INTERFACE
 *
 *
 */
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/smp_lock.h>

#include <net/udp.h>

#include <pthread.h>
#include <time.h>
#include <unistd.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>


#include "onetd.h"

MODULE_AUTHOR("Pierre Morel <pmorel@mnis.fr>");
MODULE_DESCRIPTION("Ocera Network");
MODULE_LICENSE("GPL");

/*
 * Stream side definitions
 */

char * OcnetUDPSendName = "OcnetUDPSend" ;
ocs_t         *ocn_ost_p;

/*
 * Linux thread definitions
 */
static int OceraNetworkReceiveDaemon(void *);
static int OceraNetworkSendDaemon(void *);
static int OceraNetworkCmdDaemon(void *);
static void ocnet_send_process( ocn_t *, ocera_msgbuf_t * );
static void ocnet_process(ocera_msgbuf_t * );
static struct task_struct *S_thread;
static struct semaphore	S_sem;
/*
 * OcnetCreateSocket
 *
 * CALLED BY: process (onetd)
 *
 * PURPOSE: Handle a Create request detected by the onetd daemon.
 *
 * DESCRIPTION:
 * The down stream must have been initialized by the RT thread
 * 	- first verify that it is initialized.
 *	- register the up-stream as a Linux thread
 *	- bind the two streams
 *	- create the Linux socket
 *	- Start the receive daemon
 *	- lock on semaphore
 *	- Start the send daemon
 *	- lock on semaphore
 */

int OcnetCreateSocket( ocn_t * sp){
	int error;
	unsigned long	flags;
	char name[14];
	
	ODEBUGP(ODEB_CALL,("OcnetCreateSocket: %x\n",sp));
	/* s_stream_down must have been initialized by the RT task	*/
	if ( sp->s_stream_down == NULL){
		ODEBUGP(ODEB_CALL,("OcnetCreateSocket: No stream down\n"));
		return(-1);
	}
	sprintf(name,"ocn_lx[%d]",sp->s_sock);
	/* register the up stream: the Linux side	*/
	if ( (sp->s_stream_up=ocs_register(name,ST_LINUX_TYPE)) == NULL ) {
		ODEBUGP(ODEB_CALL,("OcnetCreateSocket: cannot register stream\n"));
		return(-1);
		}
	/* bind the two streams	*/
	sp->s_stream_down->s_up = sp->s_stream_up ;
	sp->s_stream_up->s_down = sp->s_stream_down ;
	/* Initialize the semaphores	*/
	init_MUTEX_LOCKED(&sp->s_sem_down);
	init_MUTEX_LOCKED(&sp->s_sem_up);

	/* create a Linux socket		*/
	if ( (error = sock_create(PF_INET,SOCK_DGRAM,IPPROTO_UDP,&sp->s_socket)) < 0 ) {
		ODEBUGP(ODEB_CALL,("OcnetCreateSocket: Error creating socket: %d\n",error));
		rtl_spin_lock_irqsave(&ocera_socket_lock,flags);
		sp->s_state = OCN_SOCK_DEAD;
		sp->s_socket = (struct socket *) 0 ;
		ocs_unregister(sp->s_stream_up);
		rtl_spin_unlock_irqrestore(&ocera_socket_lock,flags);
		return(error);
	}
	ODEBUGP(ODEB_CALL,("OcnetCreateSocket: socket created: %p.\n",sp->s_socket));

	ODEBUGP(ODEB_CALL,("OcnetCreateSocket: send task init.\n"));
	if( ( error = kernel_thread(OceraNetworkSendDaemon,(void *)sp, CLONE_FS | CLONE_FILES | CLONE_SIGNAL | SIGCHLD)) < 0 ) {
		ODEBUGP(ODEB_CALL,("OcnetCreateSocket: ERROR on kernel_thread for OceraNetworkSendDaemon %d\n",error));
		/* TBD:  CANCEL LINUX THREAD	*/
		return(error);
	}
	ODEBUGP(ODEB_CALL,("OcnetCreateSocket: Wait on init end.\n"));
	down_interruptible(&sp->s_sem_up);
	ODEBUGP(ODEB_CALL,("OcnetCreateSocket: task init pid=%d.\n",error));
	if ( sp->s_thread_up == NULL ) {
		ODEBUGP(ODEB_CALL,("OcnetCreateSocket: null thread up END\n"));
		/* TBD:  CANCEL LINUX THREAD	*/
		return(-1);
	}
	return(0);
}

/*
 * OcnetBindSocket
 *
 * CALLED BY: process (onetd)
 *
 * PURPOSE: Handle a bind request detected by the onetd daemon.
 *
 * DESCRIPTION:
 * The down stream must have been initialized by the RT thread
 * 	- first verify that it is initialized.
 *	- Bind the Linux socket
 */

int OcnetBindSocket( ocn_t * sp){
	int error;
	int len;
	
	ODEBUGP(ODEB_CALL,("OcnetBindSocket: %x\n",sp));
	/* s_stream_down must have been initialized by the RT task	*/
	if ( sp->s_stream_down == NULL){
		ODEBUGP(ODEB_CALL,("OcnetBindSocket: No stream down\n"));
		return(-1);
	}
	if ( sp->s_stream_up == NULL){
		ODEBUGP(ODEB_CALL,("OcnetBindSocket: No stream up\n"));
		return(-1);
	}

	ODEBUGP(ODEB_CALL,("OcnetBindSocket: %x to IP %04x port %d\n",sp,sp->s_sin.sin_addr.s_addr,sp->s_sin.sin_port));
	if ( (error = sp->s_socket->ops->bind(sp->s_socket,(struct sockaddr*)&sp->s_sin,sizeof(struct sockaddr)) ) < 0 ) {
		ODEBUGP(ODEB_CALL,("OcnetBindSocket: Error binding socket: %d.\n",error));
		return(error);
	}

	len=sizeof(struct sockaddr);
	if ( (error = sp->s_socket->ops->getname(sp->s_socket,(struct sockaddr*)&sp->s_sin,&len,0) ) < 0 ) {
		ODEBUGP(ODEB_CALL,("OcnetBindSocket: on getname: %d.\n",error));
		return(error);
	}
	ODEBUGP(ODEB_CALL,("OcnetBindSocket: receive task init.\n"));


	if( ( error = kernel_thread(OceraNetworkReceiveDaemon,(void *)sp, CLONE_FS | CLONE_FILES | CLONE_SIGNAL | SIGCHLD)) < 0 ) {
		ODEBUGP(ODEB_CALL,("OcnetBindSocket: ERROR on kernel_thread for OceraNetworkReceiveDaemon %d\n",error));
		return(error);
	}

	ODEBUGP(ODEB_CALL,("OcnetBindSocket: Wait on init end.\n"));
	down_interruptible(&sp->s_sem_down);
	ODEBUGP(ODEB_CALL,("OcnetBindSocket: receive task is initialized?\n"));
	if ( sp->s_thread_down == NULL ) {
			/* sp->s_thread_down is initialized by the Linux thread
			 * This means the thread could not start	*/
		ODEBUGP(ODEB_CALL,("OcnetBindSocket: null thread down END\n"));
		/* TBD:  CANCEL LINUX THREAD	*/
		return(-1);
	}

	ODEBUGP(ODEB_CALL,("OcnetBindSocket: END\n"));
	return(0);
}

/*
 * OcnetUnbindSocket
 */
int OcnetDeleteSocket( ocn_t * sp ){
	int error=0;

	ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: Socket %d state %d\n",sp->s_sock,sp->s_state));

	if ( sp->s_thread_down != NULL ) {
		ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: Kill Receive daemon %d\n",sp->s_thread_down->pid));
		kill_proc(sp->s_thread_down->pid,SIGHUP,1);
		ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: Wait for semaphore\n"));
		down_interruptible(&sp->s_sem_down);
	}
	sp->s_thread_down = NULL;
	

	ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: releasing socket 0x%p.\n",sp->s_socket));
	if ( sp->s_socket== NULL || sp->s_socket->ops == NULL ) {
		ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: NULL socket ops return\n"));
		return(-1);
	}
	if ( (error = sp->s_socket->ops->release(sp->s_socket) ) < 0 ) {
		ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: Error releasing socket: %d.\n",error));
		return(error);
	}

	if ( sp->s_thread_up != NULL ) {
		ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: Kill Send daemon %d\n",sp->s_thread_up->pid));
		kill_proc(sp->s_thread_up->pid,SIGHUP,1);
		ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: Wait for semaphore\n"));
		down_interruptible(&sp->s_sem_up);
	}

	sp->s_thread_up = NULL;

	ocs_unregister(sp->s_stream_up);
	ocn_reset_socket(sp);
	ODEBUGP(ODEB_CALL,("OcnetDeleteSocket: END.\n"));
	return(0);
}

int OcnetUnbindSocket( ocn_t * sp ){

	ODEBUGP(ODEB_CALL,("OcnetUnbindSocket: sp = %p\n",sp));
	if ( sp->s_state == OCN_SOCK_UNBIND_WAIT)
			return(-1);
	sp->s_state = OCN_SOCK_UNBIND_WAIT;
	if ( sp->s_socket == NULL ) {
		ODEBUGP(ODEB_CALL,("OcnetUnbindSocket: NULL socket\n"));
		return(-1);
	}
	ODEBUGP(ODEB_CALL,("OcnetUnbindSocket: sp->socket = %p\n",sp->s_socket));
	if ( sp->s_socket== NULL || sp->s_socket->ops == NULL ) {
		ODEBUGP(ODEB_CALL,("OcnetUnbindSocket: NULL socket ops\n"));
		return(-1);
	}
	if ( sp->s_thread_down != NULL ) {
		ODEBUGP(ODEB_CALL,("OcnetUnbindSocket: Kill Receive daemon %d\n",sp->s_thread_down->pid));
		kill_proc(sp->s_thread_down->pid,SIGHUP,1);
		ODEBUGP(ODEB_CALL,("OcnetUnbindSocket: Wait on semaphore\n"));
		down_interruptible(&sp->s_sem_down);
	}
	sp->s_thread_down = NULL;
	ODEBUGP(ODEB_CALL,("OcnetUnbindSocket: END\n"));
	return(0);
}

/*
 * OcnetConnectSocket
 */

int OcnetConnectSocket( ocn_t * sp){
	int error;

	ODEBUGP(ODEB_CALL,("OcnetConnectSocket: %x\n",sp));
	if ( (error = sp->s_socket->ops->connect(sp->s_socket,(struct sockaddr*)&sp->s_remote,sizeof(struct sockaddr), 0) ) < 0 ) {
		ODEBUGP(ODEB_CALL,("onetd: Error connecting socket: %d.\n",error));
		return(error);
		}
	ODEBUGP(ODEB_CALL,("OcnetConnectSocket: END\n"));
	return(0);
}

/*
 * OcnetReceiveUDP
 */

int OcnetReceiveUDP(ocn_t * sp, char *data , struct sockaddr_in *from, int len ){
	int err;
	struct msghdr   msg;
	struct iovec iov;
	mm_segment_t    oldfs;

	ODEBUGP(ODEB_CALL,("OcnetReceiveUDP: on socket address %p\n",sp));
	msg.msg_name     =  from ;
	msg.msg_namelen  = sizeof ( struct sockaddr_in);
	msg.msg_iov  = &iov;
	msg.msg_iovlen   = 1;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = MSG_NOSIGNAL|MSG_WAITALL;
	msg.msg_iov->iov_len =  (__kernel_size_t) len;
	msg.msg_iov->iov_base = (char*) data;
	oldfs = get_fs(); set_fs(KERNEL_DS);
	err=sock_recvmsg(sp->s_socket,&msg,len,0);
	set_fs(oldfs);
	ODEBUGP(ODEB_CALL,("OcnetReceiveUDP: END: %d\n",err));
	return(err);
}
/*
 * OcnetSendUDP
 */
int OcnetSendUDP(ocn_t * sp,struct sockaddr_in *to, char *data , int len ){
	int err;
	struct msghdr   msg;
	struct iovec iov;
	mm_segment_t    oldfs;
	ODEBUGP(ODEB_CALL,( "OcnetSendUDP: %x\n",sp));

	if (to){
		msg.msg_name     =  to;
		msg.msg_namelen  = sizeof(struct sockaddr_in);
	}
	else {
		msg.msg_name     = 0;
		msg.msg_namelen  = 0;
	}
	msg.msg_iov  = &iov;
	msg.msg_iovlen   = 1;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = MSG_NOSIGNAL|MSG_WAITALL;
	msg.msg_iov->iov_len =  (__kernel_size_t) len;
	msg.msg_iov->iov_base = data;
	oldfs = get_fs(); set_fs(KERNEL_DS);
	err=sock_sendmsg(sp->s_socket,&msg,len);
	set_fs(oldfs);
	ODEBUGP(ODEB_CALL,( "OcnetSendUDP: END : %d\n",err));
	return(err);
}

static int OceraNetworkSendDaemon(void *arg){
	ocn_t	*sp = (ocn_t *) arg;
	ocera_msgbuf_t *mbp;
	int		sig;
	int sock;


	ODEBUGP(ODEB_CALL,("S_onetd - arg = %p\n",sp));
	if ( sp == NULL){
		ODEBUGP(ODEB_CALL,("S_onetd - null socket\n"));
		return(-1);
	}
	sock = sp->s_sock;
	if ( sp->s_stream_up == NULL ) {
		ODEBUGP(ODEB_CALL,("S_onetd - null stream\n"));
		return(-1);
	}
	if ( sp->s_stream_up->s_name == NULL ) {
		ODEBUGP(ODEB_CALL,("S_onetd - null name\n"));
		return(-1);
	}
	strcpy(current->comm,sp->s_stream_up->s_name);
	sp->s_thread_up = current;
	ODEBUGP(ODEB_CALL,("S_onetd Send goes daemonize as %d\n",current->pid));
	daemonize();
	/* Block signals	*/
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGCHLD) | sigmask(SIGSTOP)| sigmask(SIGHUP));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	/* Unlock semaphore	*/
	up(&sp->s_sem_up);

	while(1) {
		ODEBUGP(ODEB_ALL,("S_onetd -%d- Send main loop\n",sock));
		if ( ocera_net_state == OCERA_NET_STOPPED ) {
			ODEBUGP(ODEB_CALL,("S_onetd -%d- Send main loop END\n",sock));
			up(&sp->s_sem_up);
			return(0);
			}
		if ( (sig=signal_pending(current)) !=0 )  {
			ODEBUGP(ODEB_CALL,("S_onetd -%d- Send signal received %d\n",sock,sig));
			if(sp->s_state & OCN_SOCK_CLOSING){ up(&sp->s_sem_up); return(0);}
			OcnetUnbindSocket(sp);
			return(0);
		}
		ODEBUGP(ODEB_ALL,("S_onetd -%d- Send get a mbp from s_queue_up\n",sock));
		if ((mbp = ocs_get_q(&sp->s_stream_up->s_queue_up,OCS_WAIT))!=(ocera_msgbuf_t *)0){
			ODEBUGP(ODEB_ALL,("S_onetd -%d- Send got a mbp from s_queue_up\n",sock));
			if ( (sig=signal_pending(current)) !=0 )  {
				ODEBUGP(ODEB_CALL,("S_onetd -%d- Send signal received %d\n",sock,sig));
				if(sp->s_state & OCN_SOCK_CLOSING){ up(&sp->s_sem_up); return(0);}
				OcnetUnbindSocket(sp);
				return(0);
			}
			ocnet_send_process(sp,mbp);
		}
	}

	ODEBUGP(ODEB_CALL,("S_onetd: -%d- Send Daemon has ended try to release all\n",sock));
	OcnetUnbindSocket(sp);
	up(&sp->s_sem_up);
	(void)printk(KERN_NOTICE "onetd: -%d- Send Daemon end.\n",sock);
	return 0;
}

static int OceraNetworkCmdDaemon(void *sp){
	ocera_msgbuf_t *mbp;
	int		sig;

	ODEBUGP(ODEB_ALL,("onetd start\n"));
	strcpy(current->comm,"onetd");
	S_thread = current;
	ODEBUGP(ODEB_ALL,("onetd goes daemonize as %d\n",current->pid));
	daemonize();
	/* Block signals	*/
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGCHLD) | sigmask(SIGSTOP)| sigmask(SIGHUP));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	while(1) {
		ODEBUGP(ODEB_ALL,("onetd - main loop\n"));
		if ( ocera_net_state == OCERA_NET_STOPPED ) {
			ODEBUGP(ODEB_CALL,("onetd - main loop END\n"));
			up(&S_sem);
			return(0);
			}
		if ( (sig=signal_pending(current)) !=0 )  {
			ODEBUGP(ODEB_CALL,("onetd - signal received %d\n",sig));
			up(&S_sem);
			return(0);
		}
		ODEBUGP(ODEB_ALL,("onetd - get a mbp from s_queue_up\n"));
		if ((mbp = ocs_get_q(&ocn_ost_p->s_queue_up,OCS_WAIT))!=(ocera_msgbuf_t *)0){
			if ( (sig=signal_pending(current)) !=0 )  {
				ODEBUGP(ODEB_CALL,("onetd - signal received %d\n",sig));
				up(&S_sem);
				return(0);
			}
			ocnet_process(mbp);
		}
	}

	ODEBUGP(ODEB_CALL,("onetd: Send Daemon has ended try to release all\n"));
	up(&S_sem);
	(void)printk(KERN_NOTICE "onetd: Send Daemon end.\n");
	return 0;
}

static int OceraNetworkReceiveDaemon(void * arg){
	ocn_t		*sp;
	struct sock			*sk;
	ocera_msgbuf_t 		*mbp;
	char				*data;
	int					len;
	int					sig;
	int					err=0;
	struct	sockaddr_in	*remote;
	int					sock;

	ODEBUGP(ODEB_CALL,("R_onetd - arg = %p\n",arg));
	if ( arg == NULL){
		ODEBUGP(ODEB_CALL,("R_onetd - null socket\n"));
		return(-1);
	}
	sp = (ocn_t *) arg;
	if ( sp->s_stream_down == NULL ) {
		ODEBUGP(ODEB_CALL,("R_onetd - null stream down\n"));
		return(-1);
	}
	sock = sp->s_sock;
	if ( sp->s_stream_up == NULL ) {
		ODEBUGP(ODEB_CALL,("R_onetd - null stream up\n"));
		return(-1);
	}
	if ( sp->s_stream_up->s_name == NULL ) {
		ODEBUGP(ODEB_CALL,("R_onetd - null name\n"));
		return(-1);
	}

	strcpy(current->comm,sp->s_stream_down->s_name);
	sp->s_thread_down = current;
	ODEBUGP(ODEB_CALL,("R_onetd daemonize as pid: %d\n",current->pid));
	daemonize();
	/* Block signals	*/
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGCHLD) | sigmask(SIGSTOP)| sigmask(SIGHUP));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	/* Unlock semaphore	*/
	up(&sp->s_sem_down);
	sk = sp->s_socket->sk;
	while(1) {
		ODEBUGP(ODEB_CALL,("R_onetd -%d- main loop(while)\n",sock));
		if ( ocera_net_state == OCERA_NET_STOPPED ) {
			ODEBUGP(ODEB_CALL,("R_onetd -%d- main loop END\n",sock));
			if(sp->s_state & OCN_SOCK_CLOSING){
				up(&sp->s_sem_down);
				return(0);
			}
			OcnetUnbindSocket(sp);
			return(-1);
			}
		if ( (sig=signal_pending(current)) !=0 )  {
			ODEBUGP(ODEB_CALL,("R_onetd -%d signal received %d\n",sock,sig));
			if(sp->s_state & OCN_SOCK_CLOSING){
				up(&sp->s_sem_down);
				return(0);
			}
			OcnetUnbindSocket(sp);
			return(-1);
		} 
		if ( (mbp=ocs_get_q(&sp->s_stream_up->s_queue_wait,OCS_WAIT)) != 0) {
			ODEBUGP(ODEB_CALL,("R_onetd -%d- got mbp %p\n",sock,mbp));
			data = mbp->mb_datap;
			len = mbp->mb_count;
			remote =  &((struct ocn_cmd_st *)(mbp->mb_data))->o_sin;
			if ( (err=OcnetReceiveUDP(sp, data , remote , len )) >= 0 ) {
				((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_DATA_IND;
				ODEBUGP(ODEB_CALL,("R_onetd -%d- Receive: %d\n",sock,err));
				if ( sp->s_stream_down == (ocs_t *)0 ) {
					ODEBUGP(ODEB_CALL,("R_onetd -%d- stream desapear!!!\n",sock));
					OcnetUnbindSocket(sp);
					return(0);
				} else {
					ODEBUGP(ODEB_CALL,("R_onetd -%d- Data from IP 0x%08x port %d\n",sock,remote->sin_addr.s_addr, remote->sin_port));
					ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
				}
			} else {
				ODEBUGP(ODEB_CALL,("R_onetd -%d-Receive error: %d\n",sock,err));
				if ( err < 128 ) {
					ODEBUGP(ODEB_CALL,("R_onetd Raise s_sem_down\n"));
					up(&sp->s_sem_down);
					return(0);
				}
				ODEBUGP(ODEB_CALL,("R_onetd -%d- Received error: %d\n",sock,err));
				if ( (sig=signal_pending(current)) !=0 )  {
					ODEBUGP(ODEB_CALL,("R_onetd -%d- signal received %d\n",sock,sig));
					if(sp->s_state & OCN_SOCK_CLOSING){
						up(&sp->s_sem_down);
						return(0);
					}
					OcnetUnbindSocket(sp);
					return(0);
				}
				ODEBUGP(ODEB_CALL,("R_onetd -%d- Receive error: %d\n",sock,err));
				if ( sp->s_stream_up){
					ocs_put_q(&sp->s_stream_up->s_queue_wait,mbp);
				}
				else{
					return(0);
				}
			}
		} else {
			if ( (sig=signal_pending(current)) !=0 )  {
				ODEBUGP(ODEB_CALL,("R_onetd -%d- signal received %d\n",sock,sig));
				if(sp->s_state & OCN_SOCK_CLOSING){
					up(&sp->s_sem_down);
					return(0);
				}
				OcnetUnbindSocket(sp);
				up(&sp->s_sem_down);
				return(0);
			}
			ODEBUGP(ODEB_CALL,("R_onetd -%d- NO BUFFER!\n",sock));
		}
	}

	ODEBUGP(ODEB_CALL,("R_onetd -%d- END!\n",sock));
	OcnetUnbindSocket(sp);
	up(&sp->s_sem_down);
	return 0;
}


int init_module(void)
{
	int pid;

	printk( KERN_INFO "Ocera Network ("VERSION") " );
	printk( "(C)2004 Pierre Morel - MNIS\n" );

	ocera_net_state = OCERA_NET_LOADED ;

	/*
	 * Initialize PROC entries
	 */

	create_proc_read_entry("state", 0, proc_ocn, proc_ocera_network_read_state, NULL);

	/*
	 * Initialize Sockets and register Stream
	 */
	ODEBUGP(ODEB_CALL,("onetd: socket init.\n"));
	ocn_socket_init();

	ODEBUGP(ODEB_CALL,("onetd: streaminit.\n"));
	ocn_ost_p=ocs_register("RT_onetd",ST_HEADER_TYPE|ST_LINUX_TYPE);
	if ( ocn_ost_p == (ocs_t * ) 0 ) {
		printk( KERN_INFO "ocera network (version "VERSION ") unable to allocate stream table.\n");
		return(-1);
	}

	/*
	 * Initialize Send Tasks and semaphore
	 */

	ODEBUGP(ODEB_CALL,("onetd: control task init.\n"));
	init_MUTEX_LOCKED(&S_sem);
	pid=kernel_thread(OceraNetworkCmdDaemon,NULL, CLONE_FS | CLONE_FILES | CLONE_SIGNAL | SIGCHLD);
	if (pid < 0) {
		ODEBUGP(ODEB_CALL,("onetd: control task init error=%d.\n",pid));
		return(-1);
	}
	ODEBUGP(ODEB_CALL,("onetd: control task init pid=%d.\n",pid));
	/*
	 * Unlock Onetd
	 */
	ocera_net_state = OCERA_NET_STARTED ;
	/*
	 * Say hello
	 */
	printk( KERN_INFO "Ocera Network sucessfully loaded.\n");
	current->need_resched = 1;
	return(0);
}

void cleanup_module( void )
{
	ODEBUGP(ODEB_CALL,("onetd: cleanup.\n"));
	ocera_net_state = OCERA_NET_STOPPED ;
	ocn_release_all();
	ODEBUGP(ODEB_CALL,("onetd: cleanup ocs_unregister.\n"));
	ocs_unregister(ocn_ost_p);
	ODEBUGP(ODEB_CALL,("onetd: cleanup kill %d.\n",S_thread->pid));
	kill_proc(S_thread->pid,SIGHUP,1);
	ODEBUGP(ODEB_CALL,("onetd: cleanup wait on semaphore.\n"));
	down_interruptible(&S_sem);
	ODEBUGP(ODEB_CALL,("onetd: cleanup kill 2.\n"));
	kill_proc(2, SIGCHLD, 1);
	printk( KERN_INFO "Ocera Network unloaded.\n" );
	return;
}

void ocn_release_all(void){
	int 			i;
	ocn_t 			*sp;
	ocs_t			*s;

	ODEBUGP(ODEB_CALL,("ocnet_release_all\n" ));

	ocera_net_state = OCERA_NET_STOPPED ;
	remove_proc_entry("sockets",proc_ocn);
	remove_proc_entry("onetd",proc_ocera);
	for(i=0;i<OCERA_MAX_SOCKET;i++){
		sp=&ocera_socket[i];
		switch(sp->s_state){
			case OCN_SOCK_FREE:
			case OCN_SOCK_IDLE:
				break;
			case OCN_SOCK_CLOSING:
			case OCN_SOCK_DELETING:
				ODEBUGP(ODEB_CALL,("ocnet_release_all: socket: %d state: %d\n",i,sp->s_state ));
				OcnetDeleteSocket(sp);
				break;
			default:
				ODEBUGP(ODEB_CALL,("ocnet_release_all: socket: %d state: %d\n",i,sp->s_state ));
				s=sp->s_stream_down;
				OcnetDeleteSocket(sp);
				ocs_unregister(s);
				break;
		}
	}
	ODEBUGP(ODEB_CALL,("ocnet_release_all: END.\n" ));
}

static void ocnet_send_process( ocn_t *sp, ocera_msgbuf_t * mbp){
	unsigned char	cmd;
	int				nb=0;
	int				len=0;

	if ( ocera_net_state != OCERA_NET_STARTED ) {
		ODEBUGP(ODEB_CALL,("ocnet_send_process: BAD STATE\n" ));
		return;
	}
	ODEBUGP(ODEB_CALL,("ocnet_send_process: mbp: %p\n",mbp ));
	if ( mbp == NULL ){
		ODEBUGP(ODEB_CALL,("ocnet_send_process: mbp NULL\n" ));
		return;
	}
	cmd 	= ((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd ;
	ODEBUGP(ODEB_CALL,("ocnet_send_process: cmd: 0x%x state 0x%x on %p.\n",(int)cmd,sp->s_state,sp ));

	switch ( cmd ) {
	case OCN_UNBIND_REQ :
		if ( OcnetUnbindSocket(sp)){
			ODEBUGP(ODEB_CALL,("ocnet_process: Unind error.\n"));
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_UNBIND_REJ;
			sp->s_state = OCN_SOCK_DEAD;
		} else {
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_UNBIND_ACK;
			ODEBUGP(ODEB_CALL,("ocnet_process: Bind ok.\n"));
			sp->s_state = OCN_SOCK_CREATED;
		}
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	case OCN_BIND_REQ :
		if ( OcnetBindSocket(sp) ) {
			ODEBUGP(ODEB_CALL,("ocnet_process: Bind error.\n"));
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_BIND_REJ;
		} else {
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_BIND_ACK;
			ODEBUGP(ODEB_CALL,("ocnet_process: Bind ok.\n"));
			sp->s_state = OCN_SOCK_BOUND;
		}
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	case OCN_CONN_REQ :
		if ( OcnetConnectSocket(sp) ) {
			ODEBUGP(ODEB_CALL,("ocnet_process: Connect error.\n"));
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_UNCONN_IND;
		} else {
			ODEBUGP(ODEB_CALL,("ocnet_process: Connect ok.\n"));
			sp->s_state = OCN_SOCK_DATAXFER;
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_CONN_ACK;
		}
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	case OCN_UDP_REQ :
		if ( OcnetConnectSocket(sp) ) {
			ODEBUGP(ODEB_CALL,("ocnet_process: UDPsend Connect error.\n"));
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_UDP_REJ;
		} else {
			ODEBUGP(ODEB_CALL,("ocnet_process: UDPsend Connect ok.\n"));
			if ( (nb = OcnetSendUDP(sp,0,mbp->mb_datap,(int)mbp->mb_count)) < 0 ) {
				ODEBUGP(ODEB_CALL,("ocnet_process: UDPSend error: %d.\n",nb));
				((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_UDP_REJ;
			} else {
				((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_UDP_ACK;
			}
			mbp->mb_count = (short) nb ;
		}
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	case OCN_DATA_WAIT :
		ODEBUGP(ODEB_CALL,("ocnet_process: OCN_DATA_WAIT.\n"));
		ocs_put_q(&sp->s_stream_up->s_queue_wait,mbp);
		break;
	case OCN_DATA_REQ :
		ODEBUGP(ODEB_CALL,("ocnet_process: OCN_SEND_REQ.\n"));
		if ( mbp->mb_datap == (char *) 0 ) {
			ODEBUGP(ODEB_CALL,("ocnet_process: OCN_SEND_REQ: NODATA\n"));
			OcnetUnbindSocket(sp);
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_UNBIND_IND;
		}
		else {
			if ( (nb = OcnetSendUDP(sp,&((struct ocn_cmd_st *)(mbp->mb_data))->o_sin,mbp->mb_datap,(int)mbp->mb_count)) < 0 ) {
				ODEBUGP(ODEB_CALL,("ocnet_process: Send error: %d.\n",nb));
				OcnetUnbindSocket(sp);
				((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_UNBIND_IND;
				mbp->mb_count = (short) nb ;
				ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
			} else {
				((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_DATA_ACK;
				ocs_free_mbuf(mbp);
			}
		}
		break;
	case OCN_IOCTL_REQ:
		ODEBUGP(ODEB_CALL,("ocnet_process: OCN_IOCTL_REQ.\n"));
		nb=OcnetIoctl(sp,((struct ocn_cmd_st *)(mbp->mb_data))->o_flags,mbp->mb_datap);
		((struct ocn_cmd_st *)(mbp->mb_data))->o_flags = nb;
		((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_IOCTL_ACK;
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	case OCN_GETNAME_REQ:
		ODEBUGP(ODEB_CALL,("ocnet_process: OCN_GETNAME_REQ.\n"));
		len=sizeof(struct sockaddr);
		nb=sp->s_socket->ops->getname(sp->s_socket,(struct sockaddr*)&sp->s_sin,&len,0) ;
		ODEBUGP(ODEB_CALL,("ocnet_process: OCN_GETNAME_REQ return: %d.\n",nb));
		((struct ocn_cmd_st *)(mbp->mb_data))->o_flags = nb;
		((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_GETNAME_ACK;
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	case OCN_GETOPT_REQ:
		ODEBUGP(ODEB_CALL,("ocnet_process: OCN_GETOPT_REQ.\n"));
		nb=sp->s_socket->ops->getsockopt(sp->s_socket,
				((struct ocn_cmd_st *)(mbp->mb_data))->o_flags,
				((struct ocn_cmd_st *)(mbp->mb_data))->o_parm,
				mbp->mb_datap,
				&mbp->mb_count);
		((struct ocn_cmd_st *)(mbp->mb_data))->o_flags = nb;
		((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_GETOPT_ACK;
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	case OCN_SETOPT_REQ:
		ODEBUGP(ODEB_CALL,("ocnet_process: OCN_SETOPT_REQ.\n"));
		nb=sp->s_socket->ops->setsockopt(sp->s_socket,
				((struct ocn_cmd_st *)(mbp->mb_data))->o_flags,
				((struct ocn_cmd_st *)(mbp->mb_data))->o_parm,
				mbp->mb_datap,
				mbp->mb_count);
		((struct ocn_cmd_st *)(mbp->mb_data))->o_flags = nb;
		((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_SETOPT_ACK;
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	default:
		ODEBUGP(ODEB_CALL,("ocnet_send_process: default\n"));
		break;
	}
}

static void ocnet_process( ocera_msgbuf_t * mbp){
	unsigned char 	cmd;
	unsigned char 	sock;
	ocn_t			*sp;
	ocs_t			*s;

	if ( ocera_net_state != OCERA_NET_STARTED ) {
		ODEBUGP(ODEB_CALL,("ocnet_process: BAD STATE\n" ));
		return;
	}
	ODEBUGP(ODEB_CALL,("ocnet_process: mbp: %p\n",mbp ));
	if ( mbp == NULL ){
		ODEBUGP(ODEB_CALL,("ocnet_process: mbp NULL\n" ));
		return;
	}
	cmd 	= ((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd ;
	sock 	= ((struct ocn_cmd_st *)(mbp->mb_data))->o_sock ;
	if ( sock > OCERA_MAX_SOCKET ){
		ODEBUGP(ODEB_CALL,("ocnet_process: sock out of range: %d\n",(int)sock ));
		ocs_free_mbuf(mbp);
		return;
	}
	sp	= &ocera_socket[sock];
	s=sp->s_stream_down;
	ODEBUGP(ODEB_CALL,("ocnet_process: cmd: 0x%x state 0x%x on %p.\n",(int)cmd,sp->s_state,sp ));
	switch ( cmd ) {
	case OCN_CREATE_REQ :
		if ( OcnetCreateSocket(sp) ) {
			ODEBUGP(ODEB_CALL,("ocnet_process: Creation error.\n"));
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_CREATE_REJ;
		} else {
			((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd = OCN_CREATE_ACK;
			ODEBUGP(ODEB_CALL,("ocnet_process: Creation ok.\n"));
			sp->s_state = OCN_SOCK_CREATED;
		}
		ocs_put_q(&sp->s_stream_down->s_queue_down,mbp);
		break;
	case OCN_DELETE_REQ :
		if(sp->s_state & OCN_SOCK_DELETING) {
			OcnetDeleteSocket(sp);
			ocs_free_mbuf(mbp);
		} else {
			OcnetDeleteSocket(sp);
			ocs_free_mbuf(mbp);
			ocs_unregister(s);
		}
		break;
	default:
		printk( KERN_INFO "ocera network unknown command.\n" );
	}
}
/*
 * Socket Handling: getname, getsockoption and ioctl
 */


int OcnetIoctl( ocn_t * sp, unsigned int cmd, void *arg ){
	int retval = 0;

	ODEBUGP(ODEB_CALL,("OcnetIoctl: sp: 0x%p cmd: %d arg: %x.\n",sp,cmd,arg));
	if ( (retval = sp->s_socket->ops->ioctl(sp->s_socket,cmd,(unsigned long)arg) ) < 0 ) {
		ODEBUGP(ODEB_CALL,("OcnetIoctl: error: %d.\n",retval));
	}
	return(retval);
}

