/*
 * OCERA NETWORK INTERFACE
 *
 *
 */
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/socket.h>
#include <linux/in.h>

#include <net/udp.h>

#include <pthread.h>
#include <time.h>
#include <unistd.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>


#include "onetd.h"

ocn_t ocera_socket[OCERA_MAX_SOCKET];
spinlock_t	ocera_socket_lock = SPIN_LOCK_UNLOCKED;


/* 
 * int ocn_socket(int, int , int)
 *
 * CALLED by RTLinux thread
 * PARAMETERS: See man socket
 * 	int domain , int type , int protocol
 * PURPOSE:		Create an endpoint for communication and returns a descriptor
 *	DESCRIPTION:
 *		register to the Ocera Streams with name and type RTLINUX
 *		Allocate and initialize the ocn_socket structure
 *		set the command field of the request to OCN_CREATE_REQ
 *		put the mbuf on the UP queue of the COMMAND stream
 */

int ocn_socket(int domain, int type, int protocol){
	unsigned long 	flags;
	int 			sock=0;
	ocs_t			*sp;
	ocn_t 	*p = &ocera_socket[0];
	ocera_msgbuf_t  *mbp;
	char			name[16];
	int				cmd;

	ODEBUGP(ODEB_CALL,("ocn_socket: %d - %d - %d\n",domain, type, protocol));
	if ( ocera_net_state != OCERA_NET_STARTED ) {
        ODEBUGP(ODEB_CALL,("ocn_socket: BAD STATE\n" ));
        return(-1);
    }
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL) )== (ocera_msgbuf_t *) 0 ) {
		ODEBUGP(ODEB_CALL,("ocn_socket: no buffer\n"));
		return(-1);
	}
	rtl_spin_lock_irqsave(&ocera_socket_lock,flags);
	for(sock=0;sock<OCERA_MAX_SOCKET;sock++,p=&ocera_socket[sock]){
		if ( p->s_state == OCN_SOCK_FREE ) {
			p->s_state = OCN_SOCK_CREATE_WAIT;
			break;
		}
	}
	rtl_spin_unlock_irqrestore(&ocera_socket_lock,flags);
	ODEBUGP(ODEB_CALL,("ocn_socket: got sock: %d / sock addr: %p\n",sock,p));
	if ( sock >= OCERA_MAX_SOCKET ) {
		ODEBUGP(ODEB_CALL,("ocn_socket: return no slot\n"));
		return(-1);
	}
	sprintf(name,"ocn_rt[%d]",sock);

	if ( (sp=ocs_register(name,ST_RTLINUX_TYPE)) == NULL ) {
		ODEBUGP(ODEB_CALL,("ocn_socket: cannot register stream\n"));
		return(-ENOMEM);
	}
	p->s_stream_down			= sp;
	p->s_stream_up				= 0;
	p->s_type					= type;
	p->s_protocol				= protocol;
	p->s_domain					= domain;
	p->s_flags					= 0;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_CREATE_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	ocs_put_q(&ocn_ost_p->s_queue_up,mbp);
	if ( ( mbp=ocs_get_q(&sp->s_queue_down,OCS_WAIT|OCS_RTL)) == NULL ) {
		ODEBUGP(ODEB_CALL,("ocn_socket: Aborting: releasing %d\n",sock));
		ODEBUGP(ODEB_CALL,("ocn_socket: mbp %p is lost!\n",mbp));
		ocn_release(sock);
		return(-EAGAIN);
	}
	cmd = ((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd;
	ocs_free_mbuf(mbp);
	if ( cmd == OCN_CREATE_ACK) {
		ODEBUGP(ODEB_CALL,("ocn_socket: ACKED: return %d\n",sock));
		p->s_state = OCN_SOCK_CREATED;
		return sock;
	}
	ODEBUGP(ODEB_CALL,("ocn_socket: REJECTED: releasing %d\n",sock));
	ocn_release(sock);
	return(-EIO);
}

int ocn_delete(int sock){
	ocn_t 	*p;
	ocera_msgbuf_t  *mbp;

	ODEBUGP(ODEB_CALL,("ocn_delete: sock %d\n",sock));
	p=&ocera_socket[sock];
	if(p->s_state == OCN_SOCK_FREE)
			return(0);
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL) )== (ocera_msgbuf_t *) 0 ) {
		ODEBUGP(ODEB_CALL,("ocn_delete: no buffer\n"));
		return(-1);
	}
	p->s_state |= OCN_SOCK_CLOSING ;
	ODEBUGP(ODEB_CALL,("ocn_delete: message %p\n",mbp));
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_DELETE_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	ODEBUGP(ODEB_CALL,("ocn_delete: put request\n"));
	ocs_put_q(&ocn_ost_p->s_queue_up,mbp);
	ODEBUGP(ODEB_CALL,("ocn_delete: return OK\n"));
	return(0);
}

void ocn_release(int sock){
	ocn_t 	*p;
	
	ODEBUGP(ODEB_CALL,("ocn_release: slot %d\n",sock));
	p=&ocera_socket[sock];
	ocs_unregister(p->s_stream_down);
}

int ocn_close(int sock) {
	ocn_t 	*p;
	ocs_t	*s;
	
	ODEBUGP(ODEB_CALL,("ocn_close: slot %d\n",sock));
	p=&ocera_socket[sock];
	s=p->s_stream_down;
	ODEBUGP(ODEB_CALL,("ocn_close: slot %d in state %x\n",sock,p->s_state));
	switch(p->s_state){
	case OCN_SOCK_DATAXFER:
	case OCN_SOCK_CONNECTED:
	case OCN_SOCK_BOUND:
/*			ocn_unbind(sock);	*/
	case OCN_SOCK_CREATED:
			p->s_state |= OCN_SOCK_DELETING;
			ocn_delete(sock);
	case OCN_SOCK_BIND_WAIT:
			ocn_release(sock);
	case OCN_SOCK_FREE:
	case OCN_SOCK_CLOSING:
		ODEBUGP(ODEB_CALL,("ocn_close: state: %d\n",p->s_state));
		break;
	case OCN_SOCK_DEAD:
	default:
		ODEBUGP(ODEB_CALL,("ocn_close: HELP bad state\n"));
		return(EBADFD);
		break;
	}
	return(0);
}


/*	int ocn_bind(char * name, struct sockaddr_in *s , ocera_msgbuf_t * mbp)
 *
 *	CALLED BY: RTLinux thread
 *	PARAMETERS:	a reference name, local socket address, and message buffer
 *	PURPOSE:	Bind to a local socket
 *	DESCRIPTION:
 *		register to the Ocera Streams with name and type RTLINUX
 *		Allocate and initialize the ocn_socket structure
 *		set the command field of the request to OCN_CONN_REQ
 *		put the mbuf on the UP queue of the COMMAND stream
 */

int ocn_getsockname(int sock, struct sockaddr_in *s , int len){
	ocn_t 	*p ;
	p = &ocera_socket[sock];
	if ( p->s_state & OCN_SOCK_BOUND ) {
		memcpy(s,&p->s_sin,sizeof(struct sockaddr_in));
		return(0);
	}
	return(-1);
}

int ocn_bind(int sock, struct sockaddr_in *s , int len){
	unsigned long 	flags;
	ocn_t 	*p ;
	ocera_msgbuf_t	*mbp;
	int				cmd;

	rtl_spin_lock_irqsave(&ocera_socket_lock,flags);
	p = &ocera_socket[sock];
	ODEBUGP(ODEB_CALL,("ocn_bind: %d \n",sock));
	if ( p->s_state != OCN_SOCK_CREATED ) {
        ODEBUGP(ODEB_CALL,("ocn_bind: BAD STATE\n" ));
		rtl_spin_unlock_irqrestore(&ocera_socket_lock,flags);
        return(-1);
    }
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL) )== (ocera_msgbuf_t *) 0 ) {
		ODEBUGP(ODEB_CALL,("ocn_bind: NO BUFFER\n"));
		rtl_spin_unlock_irqrestore(&ocera_socket_lock,flags);
		return(-1);
	}
	p->s_state = OCN_SOCK_BIND_WAIT;
	p->s_sin.sin_family       	= s->sin_family;
	p->s_sin.sin_addr.s_addr  	= s->sin_addr.s_addr;
	p->s_sin.sin_port         	= s->sin_port;

	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_BIND_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	rtl_spin_unlock_irqrestore(&ocera_socket_lock,flags);
	ODEBUGP(ODEB_CALL,("ocn_bind: send OCN_BIND_REQ\n"));
	ocs_put_q(&p->s_stream_up->s_queue_up,mbp);
	if ( ( mbp=ocs_get_q(&p->s_stream_down->s_queue_down,OCS_WAIT|OCS_RTL)) != NULL ) {
		cmd = ((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd;
		ocs_free_mbuf(mbp);
		if ( cmd == OCN_BIND_ACK) {
			ODEBUGP(ODEB_CALL,("ocn_bind: ACKED: %d\n",sock));
			p->s_state = OCN_SOCK_BOUND;
			return 0;
		}
		else {
			ODEBUGP(ODEB_CALL,("ocn_bind: REJECTED:  %d\n",sock));
			p->s_state = OCN_SOCK_CREATED;
			return(-2);
		}
	}
	p->s_state = OCN_SOCK_DEAD;
	ODEBUGP(ODEB_CALL,("ocn_bind: Aborting: unbind %d\n",sock));
	ODEBUGP(ODEB_CALL,("ocn_bind: mbp %p is lost!\n",mbp));
	ODEBUGP(ODEB_CALL,("ocn_bind: END\n"));
	return(-3);
}
/*
 * int ocn_unbind(int sock)
 *
 * CALLED BY: rmmod when unloading
 * PARAMETERS:	socket id
 * PURPOSE:		Unbind the socket
 * DESCRIPTION:
 * 	allocate a mbuf, initialize the request
 * 	put the request to the UP queue of the COMMAND stream
 */
int ocn_unbind(int sock){
	ocera_msgbuf_t	*mbp;
	ocn_t			*p;
	int				cmd;

	ODEBUGP(ODEB_CALL,("ocn_unbind: %d\n",sock));
	if ( ocera_net_state != OCERA_NET_STARTED ) {
        ODEBUGP(ODEB_CALL,("ocn_unbind: BAD STATE\n" ));
        return(-1);
    }
	p = &ocera_socket[sock];
	if ( p->s_state == OCN_SOCK_FREE ) {
		ODEBUGP(ODEB_CALL,("ocn_unbind: error: socket is free!!\n"));
		return(0);
	}
	if ( ! p->s_state && OCN_SOCK_BOUND ) {
		ODEBUGP(ODEB_CALL,("ocn_unbind: Ooops BAD STATEn"));
		return(-1);
	}
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL)) == 0) {
		ODEBUGP(ODEB_CALL,("ocn_unbind: error: no mbuf!!\n"));
		return(-2);
	}
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_UNBIND_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	ocs_put_q(&p->s_stream_up->s_queue_up,mbp);
	if ( ( mbp=ocs_get_q(&p->s_stream_down->s_queue_down,OCS_WAIT|OCS_RTL)) != NULL ) {
		cmd = ((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd;
		ocs_free_mbuf(mbp);
		if ( cmd == OCN_UNBIND_ACK) {
			ODEBUGP(ODEB_CALL,("ocn_unbind: ACKED: %d\n",sock));
			p->s_state = OCN_SOCK_CREATED;
			return 0;
		}
		else {
			ODEBUGP(ODEB_CALL,("ocn_unbind: REJECTED:  %d\n",sock));
			ODEBUGP(ODEB_CALL,("ocn_unbind: do not know what to do\n"));
			p->s_state = OCN_SOCK_DEAD;
			return(-2);
		}
	}
	p->s_state = OCN_SOCK_DEAD;
	ODEBUGP(ODEB_CALL,("ocn_unbind: END\n"));
	return(0);
}


/*	int ocn_connect(int sock , struct sockaddr_in *r , int len)
 *
 *	CALLED BY: RTLinux thread
 *	PARAMETERS:	socket id  remote socket address, and length
 *	PURPOSE:	Connect a TCP stream to a remote socket
 *	DESCRIPTION:
 *		set the command field of the request to OCN_CONN_REQ
 *		set the address field of the request to the remote address
 *		put the mbuf on the UP queue of the UP stream
 */
int ocn_connect(int sock , struct sockaddr_in *r , int len){
	ocn_t				*p;
	ocera_msgbuf_t		*mbp;
	int					cmd;

	ODEBUGP(ODEB_CALL,("ocn_connect: %d - %x\n",sock));
	if ( ocera_net_state != OCERA_NET_STARTED ) {
        ODEBUGP(ODEB_CALL,("ocn_connect: BAD STATE\n" ));
        return(-1);
    }
	p = &ocera_socket[sock];
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL)) == 0) {
		ODEBUGP(ODEB_CALL,("ocn_connect: error: no mbuf!!\n"));
		return(-2);
	}
	p->s_remote.sin_family       = r->sin_family;
	p->s_remote.sin_addr.s_addr  = r->sin_addr.s_addr;
	p->s_remote.sin_port         = r->sin_port;
	ODEBUGP(ODEB_CALL,("ocn_connect: to:\n"));
	OCN_DEBSOCKET(ODEB_CALL,&p->s_remote);

	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_CONN_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	ocs_put_q(&p->s_stream_up->s_queue_up,mbp);
	if ( ( mbp=ocs_get_q(&p->s_stream_down->s_queue_down,OCS_WAIT|OCS_RTL)) != NULL ) {
		cmd = ((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd;
		ocs_free_mbuf(mbp);
		if ( cmd == OCN_CONN_ACK) {
			ODEBUGP(ODEB_CALL,("ocn_connect: ACKED: %d\n",sock));
			p->s_state = OCN_SOCK_DATAXFER;
			return 0;
		}
		else {
			ODEBUGP(ODEB_CALL,("ocn_connect: 0x%x on  %d\n",cmd,sock));
			return(-2);
		}
	}
	ODEBUGP(ODEB_CALL,("ocn_connect: return: no MBP\n"));
	return(-1);
}

/*  int ocn_sendto(int, void *, size_t, int, struct sockaddr *, int)
 *
 *	CALLED BY: RTLinux thread
 *	PARAMETERS:	socket id  buffer , remote socket address
 *	PURPOSE:	Send a message on a socket
 *	DESCRIPTION:
 *		set the command field of the request to OCN_UDP_REQ
 *		set the address field of the request to the remote address
 *		put the mbuf on the UP queue of the UP stream
 */
int ocn_sendto(int sock, void *data, size_t data_len, int flags, struct sockaddr_in *remote_addr, int addr_len){
	ocn_t *p;
	ocera_msgbuf_t	*mbp;
	int				ocn_flag;

	ODEBUGP(ODEB_CALL,("ocn_sendto: %d \n",sock));
	if ( remote_addr) {
		ODEBUGP(ODEB_CALL,("ocn_sendto: %d:\n"));
		OCN_DEBSOCKET(ODEB_CALL,remote_addr);
	}
	if ( ocera_net_state != OCERA_NET_STARTED ) {
        ODEBUGP(ODEB_CALL,("ocn_sendto: BAD STATE\n" ));
        return(-1);
    }
	p = &ocera_socket[sock];
	ocn_flag = OCS_WAIT ;
	if ( flags & MSG_DONTWAIT ) ocn_flag= OCS_NOWAIT;
	if ( flags & MSG_WAITALL ) ocn_flag= OCS_WAIT;
	if ( (mbp = ocs_get_mbuf(ocn_flag|OCS_RTL)) == 0) {
		ODEBUGP(ODEB_CALL,("ocn_sendto: error: no mbuf!!\n"));
		return(-EAGAIN);
	}
	ODEBUGP(ODEB_CALL,("ocn_sendto: got mbp: 0x%p\n",mbp));
	p->s_remote.sin_family       = remote_addr->sin_family;
	p->s_remote.sin_addr.s_addr  = remote_addr->sin_addr.s_addr;
	p->s_remote.sin_port         = remote_addr->sin_port;
	p->s_flags					 = flags;

	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_DATA_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sin.sin_addr.s_addr=remote_addr->sin_addr.s_addr;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sin.sin_port=remote_addr->sin_port;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sin.sin_family=remote_addr->sin_family;
	mbp->mb_datap =  data;
	mbp->mb_count = data_len;
	ocs_put_q(&p->s_stream_up->s_queue_up,mbp);
	ODEBUGP(ODEB_CALL,("ocn_sendto: return\n"));
	return(data_len);
}

/*	int ocn_init_receive(int sock,char * data , int data_len, int flags)
 *
 *	CALLED BY: RTLinux thread
 *	PARAMETERS:	socket id and data buffer, sockaddr buffer
 *	PURPOSE:	provide a buffer to receive data
 *	DESCRIPTION:
 *		set the command field of the request to OCN_DATA_WAIT
 *		put the mbuf on the UP queue of the UP stream
 */
int ocn_init_receive(int sock,char * data , int data_len, int flags){
	ocera_msgbuf_t * mbp;
	ocn_t	*sp;
	int				ocn_flag;

	ODEBUGP(ODEB_CALL,("ocn_init_receive: on %d\n",sock));
	if ( ocera_net_state != OCERA_NET_STARTED ) {
        ODEBUGP(ODEB_CALL,("ocn_init_receive: BAD STATE\n" ));
        return(-1);
    }
	sp = &ocera_socket[sock];
	if ( sp->s_state == OCN_SOCK_FREE ) {
		ODEBUGP(ODEB_CALL,("ocn_init_receive: BAD STATE\n"));
		return(-1);
	}
	ocn_flag = OCS_WAIT ;
	if ( flags & MSG_DONTWAIT ) ocn_flag= OCS_NOWAIT;
	if ( flags & MSG_WAITALL ) ocn_flag= OCS_WAIT;
	if ( (mbp = ocs_get_mbuf(ocn_flag|OCS_RTL)) == 0) {
		ODEBUGP(ODEB_CALL,("ocn_init_receive: error: no mbuf!!\n"));
		return(EAGAIN);
	}
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_DATA_WAIT;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	ocs_put_q(&sp->s_stream_up->s_queue_up,mbp);
	ODEBUGP(ODEB_CALL,("ocn_init_receive: return\n"));
	return(0);
}

/*
 * int ocn_recvfrom(int, void *,size_t,int, struct sockaddr *,int)
 *
 *	CALLED BY: RTLinux thread
 *	PARAMETERS:	socket_id and mode (WAIT/NOWAIT)
 *	PURPOSE:	get the data received on the socket
 *	DESCRIPTION:	Calls ocs_get_q on the DOWN queue of the DOWN stream
 */
int ocn_recvfrom(int sock,void *data,size_t data_len,int flags,struct sockaddr_in * addr,int *addr_len){
	ocera_msgbuf_t	*mbp;
	ocn_t			*sp=&ocera_socket[sock];
	int				ocn_flag;
	int				len;

	ODEBUGP(ODEB_CALL,("ocn_recvfrom: on sock %d - mode 0x%x\n",sock, flags));
	if ( ocera_net_state != OCERA_NET_STARTED ) {
        ODEBUGP(ODEB_CALL,("ocn_recvfrom: BAD STATE\n" ));
        return(-EBADFD);
    }
	if( sock >= MAX_OCNET ){
        ODEBUGP(ODEB_CALL,("ocn_recvfrom: BAD SOCKET\n" ));
        return(-ENOTSOCK);
	}
	sp=&ocera_socket[sock];
	if ( ! sp->s_state & OCN_SOCK_DATAXFER ){
        ODEBUGP(ODEB_CALL,("ocn_recvfrom: BAD SOCKET STATE\n" ));
        return(-EBADF);
	}
	if ( !addr_len ||  (*addr_len < sizeof(struct sockaddr_in) ) ){
		ODEBUGP(ODEB_ERROR,("ocn_recvfrom: on sock %d bad addr_len\n",sock));
        return(-EINVAL);
	}
	ocn_flag = OCS_WAIT ;
	if ( flags & MSG_DONTWAIT ) ocn_flag= OCS_NOWAIT;
	if ( flags & MSG_WAITALL ) ocn_flag= OCS_WAIT;

	if ( ocn_flag == OCS_NOWAIT ) {
		ODEBUGP(ODEB_CALL,("ocn_recvfrom: OCS_NOWAIT\n"));
		if ( (mbp=ocs_get_q(&sp->s_stream_down->s_queue_down ,OCS_NOWAIT )) == 0 ){
			ODEBUGP(ODEB_CALL,("ocn_recvfrom: error: no data waiting!!\n"));
			return(-EAGAIN);
		}
		*addr_len=sizeof(struct sockaddr_in);
		memcpy(addr,&((struct ocn_cmd_st *)(mbp->mb_data))->o_sin,*addr_len);
		len = mbp->mb_count;
		ocs_free_mbuf(mbp);
		return(len);
	}
	ODEBUGP(ODEB_CALL,("ocn_recvfrom: OCS_WAIT\n"));
	if ( (mbp = ocs_get_mbuf(OCS_WAIT|OCS_RTL)) == 0) {
		ODEBUGP(ODEB_CALL,("ocn_recvfrom: error: no mbuf!!\n"));
		return(-EAGAIN);
	}
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_DATA_WAIT;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	mbp->mb_datap = data;
	mbp->mb_count = data_len;
	ocs_put_q(&sp->s_stream_up->s_queue_up,mbp);

	if ( (mbp=ocs_get_q(&sp->s_stream_down->s_queue_down ,OCS_WAIT|OCS_RTL )) == 0 ){
		ODEBUGP(ODEB_CALL,("ocn_recvfrom: error: no data waiting!!\n"));
		return(-EAGAIN);
	}
	ODEBUGP(ODEB_CALL,("ocn_recvfrom: from:\n"));
	OCN_DEBSOCKET(ODEB_CALL,&((struct ocn_cmd_st *)(mbp->mb_data))->o_sin);
	*addr_len=sizeof(struct sockaddr_in);
	memcpy(addr,&((struct ocn_cmd_st *)(mbp->mb_data))->o_sin,*addr_len);
	len = mbp->mb_count;
	ocs_free_mbuf(mbp);
	ODEBUGP(ODEB_CALL,("ocn_ocn_recvfrom: return\n"));
	return(len);
}

int ocn_getname(int sock, struct sockaddr *s, int *l){
	ocera_msgbuf_t	*mbp;
	ocn_t			*sp=&ocera_socket[sock];

	ODEBUGP(ODEB_CALL,("ocn_getname: sp: 0x%p\n",sp));
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL) )== (ocera_msgbuf_t *) 0 ) {
		ODEBUGP(ODEB_CALL,("ocn_getname: NO BUFFER\n"));
		return(-1);
	}
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_GETNAME_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	ODEBUGP(ODEB_CALL,("ocn_getname: send OCN_GETNAME_REQ\n"));
	ocs_put_q(&sp->s_stream_up->s_queue_up,mbp);
	if ( ( mbp=ocs_get_q(&sp->s_stream_down->s_queue_down,OCS_WAIT|OCS_RTL)) != NULL ) {
		*l=sizeof(struct sockaddr);
		memcpy(s,&sp->s_sin,sizeof(struct sockaddr));
		ODEBUGP(ODEB_CALL,("ocn_getname: ok\n"));
		ocs_free_mbuf(mbp);
	} else {
		ODEBUGP(ODEB_CALL,("ocn_getname: Error on ocs_get_q\n"));
		return(-1);
	}
	ODEBUGP(ODEB_CALL,("ocn_getname: END\n"));
	return(0);
}

int ocn_ioctl(int sock, unsigned int cmd, void * arg){
	ocera_msgbuf_t	*mbp;
	ocn_t			*sp=&ocera_socket[sock];
	int				retval;

	ODEBUGP(ODEB_CALL,("ocn_ioctl: sp: 0x%p\n",sp));
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL) )== (ocera_msgbuf_t *) 0 ) {
		ODEBUGP(ODEB_CALL,("ocn_ioctl: NO BUFFER\n"));
		return(-1);
	}
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_IOCTL_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_flags=cmd;
	mbp->mb_datap=arg;
	ODEBUGP(ODEB_CALL,("ocn_ioctl: send OCN_IOCTL_REQ\n"));
	ocs_put_q(&sp->s_stream_up->s_queue_up,mbp);
	if ( ( mbp=ocs_get_q(&sp->s_stream_down->s_queue_down,OCS_WAIT|OCS_RTL)) != NULL ) {
		retval = ((struct ocn_cmd_st *)(mbp->mb_data))->o_flags;
		ocs_free_mbuf(mbp);
		return(retval);
	} else {
		ODEBUGP(ODEB_CALL,("ocn_ioctl: Error on ocs_get_q\n"));
	}
	return(-1);
}

int ocn_getsockopt(int sock, int level, int optname, char *optval, int *optlen){
	ocera_msgbuf_t	*mbp;
	ocn_t			*sp=&ocera_socket[sock];
	int				retval;

	ODEBUGP(ODEB_CALL,("ocn_getsockopt: sp: 0x%p\n",sp));
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL) )== (ocera_msgbuf_t *) 0 ) {
		ODEBUGP(ODEB_CALL,("ocn_getsockopt: NO BUFFER\n"));
		return(-1);
	}
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_GETOPT_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_flags=level;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_parm=optname;
	mbp->mb_datap=optval;
	mbp->mb_count=*optlen;
	ODEBUGP(ODEB_CALL,("ocn_getsockopt: send OCN_GETOPT_REQ\n"));
	ocs_put_q(&sp->s_stream_up->s_queue_up,mbp);
	if ( ( mbp=ocs_get_q(&sp->s_stream_down->s_queue_down,OCS_WAIT|OCS_RTL)) != NULL ) {
		retval = ((struct ocn_cmd_st *)(mbp->mb_data))->o_flags;
		ocs_free_mbuf(mbp);
		return(retval);
	}
	ODEBUGP(ODEB_CALL,("ocn_getsockopt: Error on ocs_get_q\n"));
	return(-1);
}

int ocn_setsockopt(int sock, int level, int optname, char *optval, int optlen){
	ocera_msgbuf_t	*mbp;
	ocn_t			*sp=&ocera_socket[sock];
	int				retval;

	ODEBUGP(ODEB_CALL,("ocn_setsockopt: sp: 0x%p\n",sp));
	if ( (mbp = ocs_get_mbuf(OCS_NOWAIT|OCS_RTL) )== (ocera_msgbuf_t *) 0 ) {
		ODEBUGP(ODEB_CALL,("ocn_getsockopt: NO BUFFER\n"));
		return(-1);
	}
	((struct ocn_cmd_st *)(mbp->mb_data))->o_cmd=OCN_SETOPT_REQ;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_sock=sock;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_flags=level;
	((struct ocn_cmd_st *)(mbp->mb_data))->o_parm=optname;
	mbp->mb_datap=optval;
	mbp->mb_count=optlen;
	ODEBUGP(ODEB_CALL,("ocn_setsockopt: send OCN_SETOPT_REQ\n"));
	ocs_put_q(&sp->s_stream_up->s_queue_up,mbp);
	if ( ( mbp=ocs_get_q(&sp->s_stream_down->s_queue_down,OCS_WAIT|OCS_RTL)) != NULL ) {
		retval = ((struct ocn_cmd_st *)(mbp->mb_data))->o_flags;
		ocs_free_mbuf(mbp);
		return(retval);
	} else {
		ODEBUGP(ODEB_CALL,("ocn_setsockopt: Error on ocs_get_q\n"));
	}
	return(-1);
}

void ocn_showsocket(struct sockaddr_in *s){
		ODEBUGP(ODEB_DEB,("    Socket fami: %x\n", (int) s->sin_family ));
		ODEBUGP(ODEB_DEB,("    Socket addr: %x\n", (int) ntohl(s->sin_addr.s_addr) ));
		ODEBUGP(ODEB_DEB,("    Socket port: %x\n", (int) ntohs(s->sin_port) ));
}

